;Include the GameScript engine
Include "../GameScript Functions.bb"
Include "../GameScript Compiler.bb"
Include "../GameScript VM.bb"

;Setup a window
Graphics 640, 600, 0, 2
AppTitle "GameScript Example"
Cls

;Load the script
Color 0, 255, 0
Print
Print "__ Loading GameScript __________________________________________________________________________________"
Print
Print "Loading..."
Color 255, 255, 255

file$ = "example1.gs"
If CommandLine() <> "" Then file$ = CommandLine()

;A simple GSV_ThreadLoad() function call will load the script, ready to be executed. Compilation will be handled automatically
program = GSV_ThreadLoad(file)

;Check if the script loaded OK
If program <> 0 Then
	Print "Done."
Else
	Print "Failed!"
End If

;Display whatever messages the compiler has
For i = 1 To GSC_CompileMessages()
	;GSC_CompileMessageType() tells whether the message is a critical error or just a warning
	If GSC_CompileMessageType(i) = MESSAGE_WARNING Then
		Print "WARNING: " + GSC_CompileMessage(i) + "(line #" + GSC_CompileMessageLine(i) + ")"
	End If
	If GSC_CompileMessageType(i) = MESSAGE_ERROR Then
		Print "ERROR: " + GSC_CompileMessage(i) + "(line #" + GSC_CompileMessageLine(i) + ")"
	End If
Next
Print GSV_Error()

;Don't continue if there were errors
If program = 0 Then
	Print "Press any key to exit.."
	WaitKey()
	End
End If

;If there weren't errors, proceed
Color 0, 255, 0
Print
Print "__ Running GameScript __________________________________________________________________________________"
;Print "NOTE: Running in debug mode will cause a slow-down of about 10x"
Print "Program output:"
Print
Color 255, 255, 255

;Now the script is loaded and ready to run. Now, GSV_ThreadExecute() must be called for execution to begin.
;A "thread" is just another name for a script which is loaded into memory. Multiple "threads" can be loaded
;and run at the same time.
GSV_ThreadExecute(program)

;GameScript will now constantly execute the script programs within this blitz app. However, you need to
;call GSV_Update() periodically, or GameScript will not do anything.

;This loop will continue to allow GameScript to update all script programs until GSV_ThreadActive() returns
;that the program is finished running.
Repeat

	GSV_Update(100)

Until GSV_ThreadActive(program) = False

;Now just display a simple input prompt so the program won't exit before the user can see the output.
Color 128, 128, 128
Print
Print "Program has ended."
Print "Press any key to exit.."

GSV_ThreadDelete(program)

WaitKey()
End